/** \file
    \brief Demonstration of communication with USB HID class device
*/

#include "HidDevice.h"
#include "DeviceNotifier.h"

#include <iostream>

using namespace std;
using namespace nsHidDevice;

#include <windows.h>
#include <sys/timeb.h>
#include <time.h>
void DisplayLastError(char *msg)
{
    LPVOID lpMsgBuf;
    DWORD dw = GetLastError();
    if (dw)
    {
        FormatMessage(
            FORMAT_MESSAGE_ALLOCATE_BUFFER |
            FORMAT_MESSAGE_FROM_SYSTEM |
            FORMAT_MESSAGE_IGNORE_INSERTS,
            NULL,
            dw,
            MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
            (LPTSTR) &lpMsgBuf,
            0, NULL );
        cout << msg << " " << dw << ": " << (LPCTSTR)lpMsgBuf;
    }
    LocalFree(lpMsgBuf);
}

void Display(char *lpData, ...)
{
    va_list ap;
    char buf[1024]; //determines max message length

    struct timeb timebuffer;
    char *timeline;
    ftime( &timebuffer );
    timeline = ctime( & ( timebuffer.time ) );
    timeline[19]=0;
    sprintf(buf, "%8s.%03hu: ", &timeline[11], timebuffer.millitm);

    va_start(ap, lpData);
    vsnprintf(buf + strlen(buf), 1024-strlen(buf), lpData, ap);
    va_end(ap);

    cout << buf << endl;
}

int CallbackDeviceState(std::string path, bool attached)
{
    cout << "Device " << path << " ";
    if (attached)
        cout << "attached." << endl;
    else
        cout << "removed." << endl;

    return 0;
}

int main()
{
	int VendorID = 0x04D8;
	int ProductID = 0x003F;
    int status = 0;

    cout << "This is simple test program for PIC18F2550 HID device" << endl;
    cout << "running Generic HID demo from MCHPFSUSB Framework 2.5b." << endl;
    cout << "Expected VID = " << hex << VendorID << ", PID = " << ProductID << endl;
    cout << dec;

    HidDevice hidDevice;
    nsDeviceNotifier::DeviceNotifier deviceNotifier;
    nsDeviceNotifier::DeviceNotifier::pCallbackFunction = &CallbackDeviceState;


    if (hidDevice.Open(VendorID, ProductID, NULL, NULL) == 0)
    {
        cout << "Yay! Device opened, path = " << hidDevice.GetPath() << endl;

        GUID hidGuid;
        hidDevice.GetHidGuid(&hidGuid);
        status = deviceNotifier.RegisterForNotifications(hidGuid/*, hidDevice.GetHandle()*/);
        if (status)
        {
            DisplayLastError("GetLastError: ");
            cout << "Failed to register for notifications! Status = " << status << endl;
        }

        cout << "Press key and hit Enter:" << endl;
        cout << "    A - read ADC" << endl;
        cout << "    C - Close device & app" << endl;
        cout << "    D - Dump device capabilities" << endl;
        cout << "    P - get Pushbutton state" << endl;
        cout << "    T - Toggle LED" << endl;

        char sendbuf[64];
        char rcvbuf[64];

        char input = 0;
        int len;
        int val;
        while (input != 'C')
        {
            input = toupper(getchar());
            if (input == '\n' || input == '\r')
                continue;
            switch (input)
            {
            case 'A':
                sendbuf[0] = 0x37;  //request code (see PIC firmware)
                status = hidDevice.WriteReport(HidDevice::E_REPORT_OUT, 0, sendbuf, sizeof(sendbuf));
                if (status)
                    break;
                len = sizeof(rcvbuf);
                status = hidDevice.ReadReport(HidDevice::E_REPORT_IN, 0, rcvbuf, &len, 500);
                if (status)
                    break;
                val = (unsigned char)rcvbuf[1];        //LSB
                val += 256 * (unsigned char)rcvbuf[2]; //MSB
                cout << "ADC value: " << val << endl;
                break;
            case 'D':
                {
                    std::string dump;
                    status = hidDevice.DumpCapabilities(dump);
                    cout << dump;
                }
                break;
            case 'P':
                sendbuf[0] = 0x81;  //request code (see PIC firmware)
                status = hidDevice.WriteReport(HidDevice::E_REPORT_OUT, 0, sendbuf, sizeof(sendbuf));
                if (status)
                    break;
                len = sizeof(rcvbuf);
                status = hidDevice.ReadReport(HidDevice::E_REPORT_IN, 0, rcvbuf, &len, 500);
                if (status)
                    break;
                //rcvbuf[0] is an echo of the command (see PIC firmware)
                //rcvbuf[1] contains the I/O port pin value for the pushbutton.
                cout << "Pin state: " << (int)rcvbuf[1];
                if (rcvbuf[1])
                    cout << ", pushbutton not pressed" << endl;
                else
                    cout << ", pushbutton pressed" << endl;
                break;
            case 'T':
                sendbuf[0] = 0x80;
                /** \note
                The first byte is the "Report ID".  This number is used by the USB driver, but does not
                get tranmitted accross the USB bus.  The custom HID class firmware is only configured for
                one type of report, therefore, we must always initializate this byte to "0" before sending
                a data packet to the device.
                */
                status = hidDevice.WriteReport(HidDevice::E_REPORT_OUT, 0, sendbuf, sizeof(sendbuf));
                break;
            case 'C':
                break;
            default:
                cout << "Unknown command." << endl;
                break;
            }

            if (status)
            {
                DisplayLastError("GetLastError: ");
                cout << HidDevice::GetErrorDesc(status) << endl;
            }
        }
    }
    else
    {
        cout << "Device not found or failed to open" << endl;
    }
    hidDevice.Close();

    return 0;
}
